Require Import Metatheory.

Require Import models.
Require Import maps.

Module Haskell <: LANGUAGE_MODEL.
  (** Grammar of terms. *)
  Inductive trm : Set :=
  | trm_bvar          : nat -> trm
  | trm_abs           : trm -> trm
  | trm_app           : trm -> trm -> trm

  | trm_unit          : trm
  | trm_DVar          : loc -> trm
  | trm_DRgn          : rgn -> trm
  | trm_ThreadId      : tid -> trm
  | trm_return        : trm -> trm
  | trm_bind          : trm -> trm -> trm
  | trm_newDRgn       : trm
  | trm_newDVar       : trm -> trm -> trm
  | trm_readDVar      : trm -> trm
  | trm_writeDVar     : trm -> trm -> trm
  | trm_forkIO        : trm -> trm
  | trm_atomically    : trm -> trm
  | trm_inAtomically  : trm -> trm
  | trm_protected     : trm -> trm
  | trm_unprotected   : trm -> trm
  | trm_protectDRgn   : trm -> trm
  | trm_unprotectDRgn : trm -> trm
  | trm_makeRODRgn    : trm -> trm
  | trm_makeTLDRgn    : trm -> trm.

  (* Terms are stored in the heap. *)
  Definition val := trm.

  (** Opening up abstractions *)
  Fixpoint open_rec (k : nat) (M' : trm) (M : trm) {struct M} : trm :=
    match M with
      | trm_bvar i           => if k === i then M' else (trm_bvar i)
      | trm_abs M1           => trm_abs (open_rec (S k) M' M1)
      | trm_app M1 M2        => trm_app (open_rec k M' M1) (open_rec k M' M2)
      | trm_unit             => trm_unit
      | trm_DVar d           => trm_DVar d
      | trm_DRgn r           => trm_DRgn r
      | trm_ThreadId t       => trm_ThreadId t
      | trm_return M1        => trm_return (open_rec k M' M1)
      | trm_bind M1 M2       => trm_bind (open_rec k M' M1) (open_rec k M' M2)
      | trm_newDRgn          => trm_newDRgn
      | trm_newDVar M1 M2   => trm_newDVar (open_rec k M' M1) (open_rec k M' M2)
      | trm_readDVar M1      => trm_readDVar (open_rec k M' M1)
      | trm_writeDVar M1 M2  =>
                             trm_writeDVar (open_rec k M' M1) (open_rec k M' M2)
      | trm_forkIO M1        => trm_forkIO (open_rec k M' M1)
      | trm_atomically M1    => trm_atomically (open_rec k M' M1)
      | trm_inAtomically M1  => trm_inAtomically (open_rec k M' M1)
      | trm_protected M1     => trm_protected (open_rec k M' M1)
      | trm_unprotected M1   => trm_unprotected (open_rec k M' M1)
      | trm_protectDRgn M1   => trm_protectDRgn (open_rec k M' M1)
      | trm_unprotectDRgn M1 => trm_unprotectDRgn (open_rec k M' M1)
      | trm_makeRODRgn M1    => trm_makeRODRgn (open_rec k M' M1)
      | trm_makeTLDRgn M1    => trm_makeTLDRgn (open_rec k M' M1)
    end.

  Definition open M M' := open_rec 0 M' M.
  Notation "M ^^ M'" := (open M M') (at level 67).

  (* This evaluates the purely functional core of the
     language. Basically, we perform beta reduction at points where we
     need a value (e.g., when reading or writing a DVar, we need the
     name of the DVar). We could hypothetically add more features,
     like data structures or arithmetic.

     I think it would be OK to combine this with the reduce judgment
     (i.e., turn each eval rule into a reduce rule), since any
     programs where the combination would screw up will not
     typecheck. e.g.,

     newDVar newDRgn 5

     Here we don't want to evaluate newDRgn "too soon." But this won't
     typecheck anyhow (newDRgn has type DSTM DRgn, while newDVar takes
     a DRgn). Not that I have a real type system --- but I can imagine
     writing such a type system.

     I have them separate because this presentation seems much clearer
     to me.
 *)
  Inductive eval : trm -> trm -> Prop :=

  (* our core calculus only supports beta evaluation so every eval
     step will have a beta reduction *)
  | eval_beta : forall M1 M2,
    eval (trm_app (trm_abs M1) M2) (M1 ^^ M2)

  (* then we have to evaluate certain arguments - things that need to
     be evaluated to values before they can be processed *)

  | eval_newDVar : forall M1 M2 M1',
    eval M1 M1' ->
    eval (trm_newDVar M1 M2) (trm_newDVar M1' M2)

  | eval_readDVar : forall M1 M1',
    eval M1 M1' ->
    eval (trm_readDVar M1) (trm_readDVar M1')

  | eval_writeDVar : forall M1 M2 M1',
    eval M1 M1' ->
    eval (trm_writeDVar M1 M2) (trm_writeDVar M1' M2)

  | eval_protectDRgn : forall M1 M1',
    eval M1 M1' ->
    eval (trm_protectDRgn M1) (trm_protectDRgn M1')

  | eval_unprotectDRgn : forall M1 M1',
    eval M1 M1' ->
    eval (trm_unprotectDRgn M1) (trm_unprotectDRgn M1')

  | eval_makeRODRgn : forall M1 M1',
    eval M1 M1' ->
    eval (trm_makeRODRgn M1) (trm_makeRODRgn M1')

  | eval_makeTLDRgn : forall M1 M1',
    eval M1 M1' ->
    eval (trm_makeTLDRgn M1) (trm_makeTLDRgn M1').

  (* This judgment does evaluation at the monad level *)
  Inductive reduce : trm -> action trm -> trm -> option trm -> Prop :=

  (* evaluate an expression using the pure core *)
  | reduce_eval : forall M1 M1',
    eval M1 M1' ->
    reduce M1 (pure _) M1' None

  (* allocate a new DRgn *)
  | reduce_newDRgn : forall r,
    reduce trm_newDRgn (newRgn _ r) (trm_return (trm_DRgn r)) None

  (* allocate a new DVar *)
  | reduce_newDVar : forall d r M1,
    reduce (trm_newDVar (trm_DRgn r) M1) (alloc d r M1)
    (trm_return (trm_DVar d)) None

  (* read a DVar's contents *)
  | reduce_readDVar : forall d M1,
    reduce (trm_readDVar (trm_DVar d)) (read d M1) (trm_return M1) None

  (* write a DVar's contents *)
  | reduce_writeDVar : forall d M1,
    reduce (trm_writeDVar (trm_DVar d) M1) (write d M1)
    (trm_return trm_unit) None

  (* set a DRgn's protection state to protected *)
  | reduce_protectDRgn : forall r,
    reduce (trm_protectDRgn (trm_DRgn r)) (makeP _ r)
    (trm_return trm_unit) None

  (* set a DRgn's protection state to unprotected *)
  | reduce_unprotectDRgn : forall r,
    reduce (trm_unprotectDRgn (trm_DRgn r)) (makeU _ r)
    (trm_return trm_unit) None

  (* set a DRgn's protection state to read-only *)
  | reduce_makeRODRgn : forall r,
    reduce (trm_makeRODRgn (trm_DRgn r)) (makeRO _ r)
    (trm_return trm_unit) None

  (* set a DRgn's protection state to thread-local *)
  | reduce_makeTLDRgn : forall r,
    reduce (trm_makeTLDRgn (trm_DRgn r)) (makeTL _ r)
    (trm_return trm_unit) None

  (* fork a new thread *)
  | reduce_forkIO : forall M1 t,
    reduce (trm_forkIO M1) (spawn _ t) (trm_return (trm_ThreadId t)) (Some M1)

  (* begin a transaction (emit a beginP action) *)
  | reduce_atomically : forall M1,
    reduce (trm_atomically M1) (beginP _) (trm_inAtomically M1) None

  (* return from a transaction (emit an endP action) *)
  | reduce_inatomically : forall M1,
    reduce (trm_inAtomically (trm_return M1)) (endP _) (trm_return M1) None

  (* the next four rules descend into the context, under a bind,
     inAtomically, protected, or unprotected, respectively *)
  | reduce_bind_context : forall M1 M2 a M1' M_opt,
    reduce M1 a M1' M_opt ->
    reduce (trm_bind M1 M2) a (trm_bind M1' M2) M_opt
  | reduce_inAtomically_context : forall M1 a M1' M_opt,
    reduce M1 a M1' M_opt ->
    reduce (trm_inAtomically M1) a (trm_inAtomically M1') M_opt
  | reduce_protected_context : forall M1 a M1' M_opt,
    reduce M1 a M1' M_opt ->
    reduce (trm_protected M1) a (trm_protected M1') M_opt
  | reduce_unprotected_context : forall M1 a M1' M_opt,
    reduce M1 a M1' M_opt ->
    reduce (trm_unprotected M1) a (trm_unprotected M1') M_opt

  (* the last three rules return from a bind, protected or unprotected *)
  | reduce_bind : forall M1 M2,
    reduce (trm_bind (trm_return M1) M2) (pure _) (trm_app M2 M1) None
  | reduce_protected : forall M1,
    reduce (trm_protected (trm_return M1)) (pure _) (trm_return M1) None
  | reduce_unprotected : forall M1,
    reduce (trm_unprotected (trm_return M1)) (pure _) (trm_return M1) None.

  (* Program states *)
  Definition program_state := VarMap.t trm.

  (* Program steps *)
  Inductive step : program_state -> effect val -> program_state -> Prop :=
  | step_thread : forall P t M1 a M1',
    VarMap.MapsTo t M1 P ->
    reduce M1 a M1' None ->
    step P (Some (t, a)) (VarMap.add t M1' P)
  | step_spawn : forall P t1 M1 M1' M2 t2,
    VarMap.MapsTo t1 M1 P ->
    ~ VarMap.In t2 P ->
    reduce M1 (spawn _ t2) M1' (Some M2) ->
    step P (Some (t1, spawn _ t2)) (VarMap.add t2 M2 (VarMap.add t1 M1' P)).

  Definition program_step := step.

  (** Reflexive transitive closure of program step. *)
  Inductive program_step_many : program_state -> program_state -> trace val
    -> Prop :=
  | program_step_many_refl : forall P,
    program_step_many P P nil
  | program_step_many_trans : forall P1 P2 P3 tr e,
    program_step_many P1 P2 tr ->
    program_step P2 e P3 ->
    program_step_many P1 P3 (e :: tr).

  Lemma no_empty_steps : forall P P',
    ~ (program_step P None P').
  Proof.
    introv Step. inversion Step.
  Qed.

  Definition eq_program_state := (@VarMap.Equal trm).

  Lemma eq_program_state_refl : forall P,
    eq_program_state P P.
  Proof.
    intros P.
    reflexivity.
  Qed.

  Lemma eq_program_state_sym : forall P1 P2,
    eq_program_state P1 P2 ->
    eq_program_state P2 P1.
  Proof.
    intros P1 P2 Eq.
    symmetry.
    auto.
  Qed.

  Lemma eq_program_state_trans : forall P1 P2 P3,
    eq_program_state P1 P2 ->
    eq_program_state P2 P3 ->
    eq_program_state P1 P3.
  Proof.
    intros.
    transitivity P2; auto.
  Qed.

  Lemma eq_program_state_step : forall P1 e P2 P1',
    program_step P1 e P2 ->
    eq_program_state P1 P1' ->
    exists P2',
      eq_program_state P2 P2'
      /\ program_step P1' e P2'.
  Proof with eauto.
    intros P1 e P2 P1' Step PrgmEq.
    unfold eq_program_state in *.
    inversion Step; subst.

    (* thread *)
    esplit. split.
    apply VarMapFacts.add_m...
    eapply step_thread...
    rewrite <- VarMapFacts.MapsTo_m...

    (* spawn *)
    esplit. esplit.
    apply VarMapFacts.add_m...
    apply VarMapFacts.add_m...
    eapply step_spawn...
    rewrite <- VarMapFacts.MapsTo_m...
    rewrite <- VarMapFacts.In_m...
  Qed.

  Add Parametric Relation : program_state eq_program_state
    reflexivity proved by eq_program_state_refl
    symmetry proved by eq_program_state_sym
    transitivity proved by eq_program_state_trans
  as eq_program_state_rel.
End Haskell.
