Require Import haskell.
Require Import maps.
Require Import models.
Require Import types.

(* This is just a dummy type system to ensure that the language
   doesn't manufacture DVars and DRgns out of thin air. A true Haskell
   type system would require polymorphism, which, bleh. *)
Module Haskell_Well_Formed <: TYPE_MODEL.
  Module Import L := Haskell.
  Module Import HeapTyp := HeapTyping.

  (* Every well-formed term has type tt. Ill-formed terms do not
     typecheck. *)
  Definition typ := unit.

  (* This definition is not very interesting. The only rules worth
  reading are those for DVar and DRgn. *)
  Inductive typing_trm : heap_typing typ -> trm -> Prop :=
  | typing_bvar : forall HT i,
    typing_trm HT (trm_bvar i)
  | typing_abs : forall HT M1,
    typing_trm HT M1 ->
    typing_trm HT (trm_abs M1)
  | typing_app : forall HT M1 M2,
    typing_trm HT M1 ->
    typing_trm HT M2 ->
    typing_trm HT (trm_app M1 M2)
  | typing_unit : forall HT,
    typing_trm HT trm_unit
  | typing_DVar : forall PT ST d,
    VarMap.In d ST ->
    typing_trm (PT, ST) (trm_DVar d)
  | typing_DRgn : forall PT ST r,
    VarMap.In r PT ->
    typing_trm (PT, ST) (trm_DRgn r)
  | typing_ThreadId : forall HT t,
    typing_trm HT (trm_ThreadId t)
  | typing_return : forall HT M1,
    typing_trm HT M1 ->
    typing_trm HT (trm_return M1)
  | typing_bind : forall HT M1 M2,
    typing_trm HT M1 ->
    typing_trm HT M2 ->
    typing_trm HT (trm_bind M1 M2)
  | typing_newDRgn : forall HT,
    typing_trm HT trm_newDRgn
  | typing_newDVar : forall HT M1 M2,
    typing_trm HT M1 ->
    typing_trm HT M2 ->
    typing_trm HT (trm_newDVar M1 M2)
  | typing_readDVar : forall HT M1,
    typing_trm HT M1 ->
    typing_trm HT (trm_readDVar M1)
  | typing_writeDVar : forall HT M1 M2,
    typing_trm HT M1 ->
    typing_trm HT M2 ->
    typing_trm HT (trm_writeDVar M1 M2)
  | typing_forkIO : forall HT M1,
    typing_trm HT M1 ->
    typing_trm HT (trm_forkIO M1)
  | typing_atomically : forall HT M1,
    typing_trm HT M1 ->
    typing_trm HT (trm_atomically M1)
  | typing_inAtomically : forall HT M1,
    typing_trm HT M1 ->
    typing_trm HT (trm_inAtomically M1)
  | typing_protected : forall HT M1,
    typing_trm HT M1 ->
    typing_trm HT (trm_protected M1)
  | typing_unprotected : forall HT M1,
    typing_trm HT M1 ->
    typing_trm HT (trm_unprotected M1)
  | typing_protectDRgn : forall HT M1,
    typing_trm HT M1 ->
    typing_trm HT (trm_protectDRgn M1)
  | typing_unprotectDRgn : forall HT M1,
    typing_trm HT M1 ->
    typing_trm HT (trm_unprotectDRgn M1)
  | typing_makeRODRgn : forall HT M1,
    typing_trm HT M1 ->
    typing_trm HT (trm_makeRODRgn M1)
  | typing_makeTLDRgn : forall HT M1,
    typing_trm HT M1 ->
    typing_trm HT (trm_makeTLDRgn M1).

  Hint Constructors typing_trm.

  (* The vaue typing judgment just delegates to the term typing
     judgment. *)
  Inductive typing_val_ind : heap_typing typ -> val -> typ -> Prop :=
  | typing_val_def : forall HT v,
    typing_trm HT v ->
    typing_val_ind HT v tt.
  Definition typing_val := typing_val_ind.

  (* The program typing judgment checks that each term in the program
     state is well-formed. *)
  Inductive typing_prgm_ind : heap_typing typ -> program_state -> Prop :=
  | typing_prgm_def : forall HT P,
    (forall t M1,
      VarMap.MapsTo t M1 P ->
      typing_trm HT M1) ->
    typing_prgm_ind HT P.
  Definition typing_prgm := typing_prgm_ind.

  Lemma typing_trm_stability : forall HT M HT',
    typing_trm HT M ->
    HeapTyp.extends HT HT' ->
    typing_trm HT' M.
  Proof with eauto.
    intros HT M HT' TypingTrm Extends.
    induction TypingTrm as [ | | | | ? ? ? TypingDVar | ? ? ? TypingDRgn
      | | | | | | | | | | | | | | | | ]; eauto; inversion Extends; subst.

    (* DVar *)
    destruct TypingDVar.
    apply typing_DVar...

    (* DRgn *)
    destruct TypingDRgn.
    apply typing_DRgn...
  Qed.

  (* This is a standard result: extending the heap typing to include
     more DVars/DRgns will never cause a well-typed term not to
     typecheck. *)
  Fact typing_stability : forall HT v T HT',
    typing_val HT v T ->
    HeapTyp.extends HT HT' ->
    typing_val HT' v T.
  Proof with eauto.
    intros HT v T HT' TypingVal Extends.
    destruct T.
    inversion TypingVal; subst.
    apply typing_val_def.
    eapply typing_trm_stability...
  Qed.

  Inductive well_typed_heap_action_ind : HeapTyp.heap_typing typ
    -> action L.val -> Prop :=
  | well_typed_heap_action_pure : forall PT ST,
    well_typed_heap_action_ind (PT, ST) (pure _)
  | well_typed_heap_action_newRgn : forall PT ST r,
    ~ VarMap.In r PT ->
    well_typed_heap_action_ind (PT, ST) (newRgn _ r)
  | well_typed_heap_action_alloc : forall PT ST l r v,
    ~ VarMap.In l ST ->
    well_typed_heap_action_ind (PT, ST) (alloc l r v)
  | well_typed_heap_action_read : forall PT ST l v T,
    VarMap.MapsTo l T ST ->
    typing_val (PT, ST) v T ->
    well_typed_heap_action_ind (PT, ST) (read l v)
  | well_typed_heap_action_write : forall PT ST l v,
    well_typed_heap_action_ind (PT, ST) (write l v)
  | well_typed_heap_action_spawn : forall PT ST t,
    well_typed_heap_action_ind (PT, ST) (spawn _ t)
  | well_typed_heap_action_makeP : forall PT ST r,
    well_typed_heap_action_ind (PT, ST) (makeP _ r)
  | well_typed_heap_action_makeU : forall PT ST r,
    well_typed_heap_action_ind (PT, ST) (makeU _ r)
  | well_typed_heap_action_makeRO : forall PT ST r,
    well_typed_heap_action_ind (PT, ST) (makeRO _ r)
  | well_typed_heap_action_makeTL : forall PT ST r,
    well_typed_heap_action_ind (PT, ST) (makeTL _ r)
  | well_typed_heap_action_beginP : forall PT ST,
    well_typed_heap_action_ind (PT, ST) (beginP _)
  | well_typed_heap_action_endP : forall PT ST,
    well_typed_heap_action_ind (PT, ST) (endP _).

  Definition well_typed_heap_action := well_typed_heap_action_ind.

  Inductive well_typed_program_action_ind : HeapTyp.heap_typing typ
    -> action L.val -> HeapTyp.heap_typing typ -> Prop :=
  | well_typed_program_action_pure : forall PT ST,
    well_typed_program_action_ind (PT, ST) (pure _) (PT, ST)
  | well_typed_program_action_newRgn : forall PT ST r,
    well_typed_program_action_ind (PT, ST) (newRgn _ r) (VarMap.add r tt PT, ST)
  | well_typed_program_action_alloc : forall PT ST l r v T,
    VarMap.In r PT ->
    typing_val (PT, ST) v T ->
    well_typed_program_action_ind (PT, ST) (alloc l r v) (PT, VarMap.add l T ST)
  | well_typed_program_action_read : forall PT ST l v,
    well_typed_program_action_ind (PT, ST) (read l v) (PT, ST)
  | well_typed_program_action_write : forall PT ST l v T,
    VarMap.MapsTo l T ST ->
    typing_val (PT, ST) v T ->
    well_typed_program_action_ind (PT, ST) (write l v) (PT, ST)
  | well_typed_program_action_spawn : forall PT ST t,
    well_typed_program_action_ind (PT, ST) (spawn _ t) (PT, ST)
  | well_typed_program_action_makeP : forall PT ST r,
    VarMap.In r PT ->
    well_typed_program_action_ind (PT, ST) (makeP _ r) (PT, ST)
  | well_typed_program_action_makeU : forall PT ST r,
    VarMap.In r PT ->
    well_typed_program_action_ind (PT, ST) (makeU _ r) (PT, ST)
  | well_typed_program_action_makeRO : forall PT ST r,
    VarMap.In r PT ->
    well_typed_program_action_ind (PT, ST) (makeRO _ r) (PT, ST)
  | well_typed_program_action_makeTL : forall PT ST r,
    VarMap.In r PT ->
    well_typed_program_action_ind (PT, ST) (makeTL _ r) (PT, ST)
  | well_typed_program_action_beginP : forall PT ST,
    well_typed_program_action_ind (PT, ST) (beginP _) (PT, ST)
  | well_typed_program_action_endP : forall PT ST,
    well_typed_program_action_ind (PT, ST) (endP _) (PT, ST).

  Definition well_typed_program_action := well_typed_program_action_ind.

  Lemma well_typed_extends :
    forall (HT : HeapTyp.heap_typing typ) a HT',
      well_typed_heap_action HT a ->
      well_typed_program_action HT a HT' ->
      HeapTyp.extends HT HT'.
  Proof with eauto.
    intros HT a HT' TypingHeap TypingPrgm.
    inversion TypingHeap; subst;
      inversion TypingPrgm; subst;
        apply HeapTyping.extends_def...
    intros.
    rewrite VarMapFacts.add_neq_in_iff...
    intro.
    subst...
    intros.
    rewrite VarMapFacts.add_neq_mapsto_iff...
    intro.
    subst...
  Qed.
End Haskell_Well_Formed.

Ltac inversion_typing_trm TypingTrm Typing1 Typing2 :=
  inversion TypingTrm as [ | | |
    | ? ? ? Typing1 (* DVar *)
    | ? ? ? Typing1 (* DRgn *)
    | (* ThreadId *)
    | (* return *)
    | ? ? ? Typing1 Typing2 (* bind *)
    | (* newDRgn *)
    | ? ? ? Typing1 Typing2 (* newDVar *)
    | ? ? Typing1 (* readDVar *)
    | ? ? ? Typing1 Typing2 (* writeDVar *)
    | | |
    | ? ? Typing1 (* protected *)
    | ? ? Typing1 (* unprotected *)
    | ? ? Typing1 (* protectDRgn *)
    | ? ? Typing1 (* unprotectDRgn *)
    | ? ? Typing1 (* makeRODRgn *)
    | ? ? Typing1 (* makeTLDRgn *)
  ]; subst.

Ltac inversion_well_typed_heap_action TypingA FreshRgn FreshLoc AllocatedLoc
  TypingLoc :=
  inversion TypingA as
    [ (* pure *)
      | ? ? ? FreshRgn (* newRgn *)
      | ? ? ? ? ? FreshLoc (* alloc *)
      | ? ? ? ? ? AllocatedLoc TypingLoc (* read *)
      | (* write *)
      | (* spawn *)
      | (* makeP *)
      | (* makeU *)
      | (* makeRO *)
      | (* makeTL *)
      | (* beginP *)
      | (* endP *)
    ]; subst.

Ltac inversion_well_typed_program_action TypingA TypingRgn TypingVal TypingLoc
  :=
  inversion TypingA as [
    (* pure *)
    | (* newRgn *)
    | ? ? ? ? ? ? TypingRgn TypingVal (* alloc *)
    | (* read *)
    | ? ? ? ? ? TypingLoc TypingVal (* write *)
    | (* spawn *)
    | ? ? ? TypingRgn (* makeP *)
    | ? ? ? TypingRgn (* makeU *)
    | ? ? ? TypingRgn (* makeRO *)
    | ? ? ? TypingRgn (* makeTL *)
    | (* beginP *)
    | (* endP *)
  ]; subst.

Module Haskell_Well_Formed_Preservation <:
  PROGRAM_TYPE_PRESERVATION Haskell_Well_Formed.
  Import Haskell.
  Import Haskell_Well_Formed.

  Lemma type_preservation_open : forall HT k M1 M2,
    typing_trm HT M1 ->
    typing_trm HT M2 ->
    typing_trm HT (open_rec k M1 M2).
  Proof with eauto.
    intros HT k M1 M2 TypingM1 TypingM2.
    generalize dependent k.
    induction M2; intros k; simpl; try inversion TypingM2...
    destruct (Peano_dec.eq_nat_dec k n)...
  Qed.

  Lemma type_preservation_eval : forall HT M1 M2,
    typing_trm HT M1 ->
    eval M1 M2 ->
    typing_trm HT M2.
  Proof with eauto.
    intros HT M1 M2 TypingTrm Ev.
    induction Ev;
      inversion TypingTrm as
        [ | | ? ? ? TypingM2
          | | | | | | | | | | | | | | | | | | | ];
        subst...
    inversion TypingM2.
    apply type_preservation_open...
  Qed.

  Hint Constructors well_typed_program_action_ind.
  Hint Unfold well_typed_program_action typing_val.
  Hint Constructors typing_trm typing_val_ind.
  Hint Resolve type_preservation_eval.

  Lemma type_preservation_reduce : forall HT M1 a M2 M_opt,
    typing_trm HT M1 ->
    well_typed_heap_action HT a ->
    reduce M1 a M2 M_opt ->
    exists HT',
      well_typed_program_action HT a HT'
      /\ typing_trm HT' M2
      /\ match M_opt with
           | None => True
           | Some M => typing_trm HT' M
         end.
  Proof with eauto 7.
    intros [ PT ST ] M1 M2 a M_opt TypingTrm TypingAction Reduce.
    induction Reduce.

    (* eval *)
    eauto 6.

    (* newDRgn *)
    esplit. split. auto. split; auto.
    apply typing_return.
    apply typing_DRgn.
    rewrite VarMapFacts.add_in_iff...

    (* newDVar *)
    inversion_typing_trm TypingTrm TypingDRgn x.
    inversion_typing_trm TypingDRgn AllocatedDRgn y.
    inversion TypingDRgn; subst.
    esplit. split.
    apply well_typed_program_action_alloc...
    split; auto.
    apply typing_return.
    apply typing_DVar.
    rewrite VarMapFacts.add_in_iff...

    (* readDVar *)
    inversion_well_typed_heap_action TypingAction x x x TypingDVar.
    inversion TypingDVar; subst...

    (* writeDVar *)
    inversion_typing_trm TypingTrm TypingDVar x.
    inversion_typing_trm TypingDVar AllocatedDVar y.
    destruct AllocatedDVar as [ [] ].
    esplit. split...

    (* protectDRgn *)
    inversion_typing_trm TypingTrm TypingDRgn x.
    inversion TypingDRgn; subst...

    (* unprotectDRgn *)
    inversion_typing_trm TypingTrm TypingDRgn x.
    inversion TypingDRgn; subst...

    (* makeRODRgn *)
    inversion_typing_trm TypingTrm TypingDRgn x.
    inversion TypingDRgn; subst...

    (* makeTLDRgn *)
    inversion_typing_trm TypingTrm TypingDRgn x.
    inversion TypingDRgn; subst...

    (* forkIO *)
    inversion TypingTrm; subst...

    (* atomically *)
    inversion TypingTrm; subst...

    (* inAtomically *)
    inversion TypingTrm; subst...

    (* bind context *)
    inversion TypingTrm; subst.
    specializes IHReduce (>>> __ __)...
    destruct IHReduce as [ ? [ ? [ ? ? ] ] ].
    esplit. split. eauto. split; auto.
    apply typing_bind; auto.
    eapply typing_trm_stability...
    eapply well_typed_extends...

    (* inAtomically context *)
    inversion TypingTrm; subst.
    specializes IHReduce (>>> __ __)...
    destruct IHReduce as [ ? [ ? [ ? ? ] ] ]...

    (* protected context *)
    inversion TypingTrm; subst.
    specializes IHReduce (>>> __ __)...
    destruct IHReduce as [ ? [ ? [ ? ? ] ] ]...

    (* unprotected context *)
    inversion TypingTrm; subst.
    specializes IHReduce (>>> __ __)...
    destruct IHReduce as [ ? [ ? [ ? ? ] ] ]...

    (* bind *)
    inversion_typing_trm TypingTrm TypingM1 x.
    inversion TypingM1...

    (* protected *)
    inversion_typing_trm TypingTrm TypingM1 x.
    inversion TypingM1...

    (* unprotected *)
    inversion_typing_trm TypingTrm TypingM1 x.
    inversion TypingM1...
  Qed.

  Fact type_preservation_program_step_mutual : forall HT P t a P',
    typing_prgm HT P ->
    program_step P (Some (t, a)) P' ->
    well_typed_heap_action HT a ->
    exists HT',
      well_typed_program_action HT a HT'
      /\ typing_prgm HT' P'.
  Proof with eauto.
    intros HT P t a P' TypingPrgm Step TypingAction.
    inversion TypingPrgm as [ ? ? TypingThreads ]; subst.
    inversion Step; subst.

    (* non-spawn step *)
    specializes TypingThreads (>>> __ __ __)...
    forwards [ ? [ ? [ ? ? ] ] ] : type_preservation_reduce...
    forwards : well_typed_extends...
    esplit. split. eauto.
    apply typing_prgm_def.
    intros t' M2 Thread.
    rewrite VarMapFacts.add_mapsto_iff in Thread.
    destruct Thread as [ [ ? ? ] | [ ? ? ] ].
    subst...
    inversion TypingPrgm.
    eapply typing_trm_stability...

    (* spawn step *)
    specializes TypingThreads (>>> __ __ __)...
    forwards [ ? [ ? [ ? ? ] ] ] : type_preservation_reduce...
    forwards : well_typed_extends...
    esplit. split. eauto.
    apply typing_prgm_def.
    intros t' M3 Thread.
    rewrite VarMapFacts.add_mapsto_iff in Thread.
    destruct Thread as [ [ ? ? ] | [ ? Thread ] ].
    subst...
    rewrite VarMapFacts.add_mapsto_iff in Thread.
    destruct Thread as [ [ ? ? ] | [ ? ? ] ].
    subst...
    inversion TypingPrgm.
    eapply typing_trm_stability...
  Qed.
End Haskell_Well_Formed_Preservation.
