Require Import maps.
Require Import models.

Set Implicit Arguments.

Open Scope type_scope.

(* This module type gathers all of the heap structures. *)
Module Type HEAP (L : LANGUAGE_MODEL).
  Inductive tx_state :=
  | tx_none
  | tx_running : tid -> tx_state
  | tx_commit : tid -> tx_state.

  (* Map from locations to values. *)
  Definition store := VarMap.t (rgn * L.val).

  (* Map from locations to protection states. *)
  Definition prot := VarMap.t prot_state.

  (* Per-location states for the transaction log. *)
  Inductive tx_loc_state :=
  | tx_ro (* read-only *)
  | tx_mod (* modified *).

  (* Buffered values for the current transaction. *)
  Definition tx_log := VarMap.t (tx_loc_state * L.val).

  (* Inverted store (tx state + value) *)
  Definition inverted_store := VarMap.t (option tx_loc_state * rgn * L.val).

  (* The three heap forms *)
  Definition strong_heap := tx_state * store * prot.
  Definition weak_heap := tx_state * store * prot * tx_log.
  Definition inverted_heap := tx_state * inverted_store * prot.

  (* The three empty heaps *)
  Definition empty_strong_heap : strong_heap :=
    (tx_none, VarMap.empty _, VarMap.empty _).
  Definition empty_weak_heap : weak_heap :=
    (tx_none, VarMap.empty _, VarMap.empty _, VarMap.empty _).
  Definition empty_inverted_heap : inverted_heap :=
    (tx_none, VarMap.empty _, VarMap.empty _).

  Inductive strong_active : tx_state -> tid -> Prop :=
  | strong_active_none : forall t,
    strong_active tx_none t
  | strong_active_running : forall t,
    strong_active (tx_running t) t.

  Inductive weak_active : tx_state -> tid -> Prop :=
  | weak_active_none : forall t,
    weak_active tx_none t
  | weak_active_running : forall t t',
    weak_active (tx_running t) t'
  | weak_active_commit : forall t t',
    t <> t' ->
    weak_active (tx_commit t) t'.

  Inductive weak_unprotected : tx_state -> tid -> Prop :=
  | weak_unprotected_none : forall t,
    weak_unprotected tx_none t
  | weak_unprotected_running : forall t t',
    t <> t' ->
    weak_unprotected (tx_running t') t
  | weak_unprotected_commit : forall t t',
    t <> t' ->
    weak_unprotected (tx_commit t') t.

  Inductive eq_weak_heap : weak_heap -> weak_heap -> Prop :=
  | eq_weak_heap_def : forall ts S S' R R' TS TS',
    VarMap.Equal S S' ->
    VarMap.Equal R R' ->
    VarMap.Equal TS TS' ->
    eq_weak_heap (ts, S, R, TS) (ts, S', R', TS').

  Lemma eq_weak_heap_refl : forall H,
    eq_weak_heap H H.
  Proof.
    intros [ [ [ ? ? ] ? ] ? ].
    apply eq_weak_heap_def;
      reflexivity.
  Qed.

  Lemma eq_weak_heap_sym : forall H1 H2,
    eq_weak_heap H1 H2 ->
    eq_weak_heap H2 H1.
  Proof.
    intros H1 H2 Eq.
    inversion Eq; subst.
    apply eq_weak_heap_def;
      symmetry; auto.
  Qed.

  Lemma eq_weak_heap_trans : forall H1 H2 H3,
    eq_weak_heap H1 H2 ->
    eq_weak_heap H2 H3 ->
    eq_weak_heap H1 H3.
  Proof with auto.
    intros H1 H2 H3 Eq1 Eq2.
    destruct H1 as [ [ [ ts1 S1 ] R1 ] TS1 ].
    destruct H2 as [ [ [ ts2 S2 ] R2 ] TS2 ].
    destruct H3 as [ [ [ ts3 S3 ] R3 ] TS3 ].
    inversion Eq1; subst.
    inversion Eq2; subst.
    apply eq_weak_heap_def.
    transitivity S2...
    transitivity R2...
    transitivity TS2...
  Qed.

  Add Parametric Relation : weak_heap eq_weak_heap
    reflexivity proved by eq_weak_heap_refl
    symmetry proved by eq_weak_heap_sym
    transitivity proved by eq_weak_heap_trans
  as eq_weak_heap_rel.
End HEAP.
