(* Collecting useful map modules together. *)

Require Export FMaps.
Require Import Metatheory.
Require Import OrderedTypeEx.

Set Implicit Arguments.

Declare Module VarMap : S with Module E := Var_as_OT.
Module VarPair := PairOrderedType Var_as_OT Var_as_OT.
Declare Module VarPairMap : S with Module E := VarPair.
Module VarMapFacts := Facts VarMap.
Module VarPairMapFacts := Facts VarPairMap.
Module VarMapProperties := Properties VarMap.
Module VarPairMapProperties := Properties VarPairMap.

Hint Unfold VarMap.In.
Hint Unfold VarPairMap.In.

Module MapFactsEx (Map : S).
  Import Map.
  Module Import MapFacts := Facts Map.
  Module Import MapProperties := Properties Map.

  Section Types.

    Variable elt : Type.
    Implicit Type m:t elt.
    Implicit Type e:elt.

    Lemma add_eq_mapsto_iff : forall m x y e e',
      E.eq x y -> (MapsTo y e' (add x e m) <-> e = e').
    Proof.
      intros m x y e e' Eq. split.
      intros Maps.
      rewrite add_mapsto_iff in Maps.
      destruct Maps as [ [ ? ? ] | [ ? ? ] ]. auto. tauto.
      intro. subst. apply find_2.
      rewrite* add_eq_o.
    Qed.

    Lemma add_eq_in_iff : forall m x y e,
      E.eq x y -> In y (add x e m).
    Proof.
      intros m x y e Eq.
      esplit.
      rewrite* add_eq_mapsto_iff.
    Qed.

    Lemma empty_in_iff : forall x m,
      Empty m ->
      (In x m <-> False).
    Proof.
      intros x m Emp.
      unfold Empty in Emp. split.
      intros [ e Maps ].
      specializes Emp (>>> x e). tauto.
      intros false. tauto.
    Qed.

    Lemma empty_find : forall m,
      Empty m ->
      forall x, find x m = None.
    Proof.
      intros m Emp x.
      unfold Empty in Emp.
      remember (find x m) as e_opt.
      destruct e_opt as [ e | ].
      symmetry in Heqe_opt.
      apply find_2 in Heqe_opt.
      specializes Emp (>>> x e).
      tauto. auto.
    Qed.

    Lemma remove_eq_mapsto_iff : forall x y e m,
      E.eq x y ->
      (MapsTo x e (remove y m) <-> False).
    Proof.
      intros x y e m Eq.
      split. intro Maps.
      apply E.eq_sym in Eq.
      apply remove_1 with (m := m) in Eq.
      apply Eq. esplit. eauto.
      intro false. tauto.
    Qed.

    Lemma remove_not_in_iff : forall x y m,
      ~ In x (remove y m) <->
      E.eq x y \/ ~ In x m.
    Proof.
      intros x y m. split.
      intros NotIn. destruct (E.eq_dec x y). auto.
      right. intro Ins. apply NotIn.
      rewrite* remove_neq_in_iff.
      intros [ Eq | NotIn ].
      apply remove_1. auto.
      intros [ ? ? ]. apply NotIn. esplit.
      eapply remove_3. eauto.
    Qed.

    Lemma cardinal_remove : forall m x n,
      cardinal m = S n ->
      In x m ->
      cardinal (remove x m) = n.
    Proof with eauto.
      intros m x n Card [ e Maps ].
      forwards AddSize : (@MapProperties.cardinal_2 elt) (>>> __ m x e).
      apply Map.remove_1.
      apply E.eq_refl.
      intros x'. destruct (E.eq_dec x x').
      rewrite MapFacts.add_eq_o...
      apply Map.find_1.
      rewrite MapFacts.MapsTo_iff...
      rewrite MapFacts.add_neq_o...
      rewrite MapFacts.remove_neq_o...
      rewrite Card in AddSize.
      inversion AddSize...
    Qed.

    Lemma Equal_add_add : forall x y ex ey m,
      ~ E.eq x y ->
      Equal (add x ex (add y ey m)) (add y ey (add x ex m)).
    Proof.
      intros x y ex ey m Neq.
      intros z.
      repeat rewrite add_o.
      destruct (E.eq_dec x z);
        destruct (E.eq_dec y z);
          auto.
      contradiction Neq.
      eapply E.eq_trans; eauto.
    Qed.

    Lemma Equal_add_add_add : forall x y z ex ey ez m,
      ~ E.eq x z ->
      ~ E.eq y z ->
      Equal (add x ex (add y ey (add z ez m)))
      (add z ez (add x ex (add y ey m))).
    Proof.
      intros x y z ex ey ez m Neq1 Neq2.
      intros u.
      repeat rewrite add_o.
      destruct (E.eq_dec x u);
        destruct (E.eq_dec y u);
          destruct (E.eq_dec z u);
            auto.
      contradiction Neq1.
      eapply E.eq_trans; eauto.
      contradiction Neq1.
      eapply E.eq_trans; eauto.
      contradiction Neq2.
      eapply E.eq_trans; eauto.
    Qed.
  End Types.
End MapFactsEx.

Module VarMapFactsEx2 := MapFactsEx VarMap.
Module VarPairMapFactsEx2 := MapFactsEx VarPairMap.

Module VarMapFactsEx.
  Lemma add_not_in_iff : forall elt x y (e:elt) m,
    ~ VarMap.In x (VarMap.add y e m) <->
    x <> y /\ ~ VarMap.In x m.
  Proof.
    intros elt x y e m. split.
    intros NotIn. split.
      intro. subst. apply NotIn.
      esplit. apply VarMap.find_2.
      rewrite* VarMapFacts.add_eq_o.
      intro. apply NotIn.
      rewrite* VarMapFacts.add_in_iff.
    intros [ ? NotIn ] In.
      apply NotIn.
      rewrite* VarMapFacts.add_in_iff in In.
      destruct In. subst. tauto. auto.
  Qed.
End VarMapFactsEx.

Module VarPairMapFactsEx.
  Lemma add_not_in_iff : forall elt x y (e:elt) m,
    ~ VarPairMap.In x (VarPairMap.add y e m) <->
    x <> y /\ ~ VarPairMap.In x m.
  Proof.
    intros elt [ x1 x2 ] [ y1 y2 ] e m. split.
    intros NotIn. split.
      intro Eq. inversion Eq. subst. apply NotIn.
      esplit. apply VarPairMap.find_2.
      rewrite* VarPairMapFacts.add_eq_o.
      intro. apply NotIn.
      rewrite* VarPairMapFacts.add_in_iff.
    intros [ ? NotIn ] In.
      apply NotIn.
      rewrite* VarPairMapFacts.add_in_iff in In.
      destruct In as [ [ ? ? ] | ].
      simpl in *. subst. tauto. auto.
  Qed.
End VarPairMapFactsEx.
