Require Import Arith.

Require Import indexes.
Require Import list_facts.
Require Import models.
Require Import nat_facts.

Set Implicit Arguments.

Open Scope nat_scope.

Module LanguageModelFacts (L : LANGUAGE_MODEL).
  Import L.

  Hint Constructors program_step_many.

  Lemma program_step_many_alt : forall P P' P'' tr tr',
    program_step_many P P' tr ->
    program_step_many P' P'' tr' ->
    program_step_many P P'' (tr' ++ tr).
  Proof.
    intros P P' P'' tr tr' steps_many steps_many'.
    induction steps_many'.
    simpl. auto.
    rewrite <- app_comm_cons. eauto.
  Qed.

  Lemma program_step_many_index_inversion : forall P1 P2 tr j e,
    program_step_many P1 P2 tr ->
    index tr j e ->
    exists P3, exists P4, exists tr1, exists tr2,
    program_step_many P1 P3 tr1 /\
    program_step P3 e P4 /\
    program_step_many P4 P2 tr2 /\
    tr = tr2 ++ (e :: nil) ++ tr1.
  Proof.
    intros P1 P2 tr j e steps_many idx.

    (* Proof by induction on the derivation of steps_many. *)
    generalize dependent j.
    induction steps_many as [ | P1 P2 P3 tr e' steps_many ? steps ];
      intros j idx; subst.

    (* Zero steps total - vacuous. *)
    inversion idx.

    (* At least one step - was j the most recent? *)
    inversion idx as [ | ? j' ? ? idx' ]; subst.

    (* Yes - so tr' is nil and tr'' is tr. *)
    exists P2. exists P3. exists tr. exists (@nil (effect val)).
    simpl. auto.

    (* No - use the induction hypothesis. *)
    pose proof IHsteps_many _ idx'
      as [ ? [ ? [ ? [ ? [ steps_many' [ ? [ ? ? ] ] ] ] ] ] ].
    subst. repeat esplit. apply steps_many'. eauto.
    eapply program_step_many_trans; eauto. eauto.
  Qed.
End LanguageModelFacts.

Module MemoryModelFacts (L : LANGUAGE_MODEL) (M : MEMORY_MODEL L).
  Import M.

  Hint Constructors heap_step_many.

  Lemma heap_step_many_alt : forall H H' H'' tr tr',
    heap_step_many H H' tr ->
    heap_step_many H' H'' tr' ->
    heap_step_many H H'' (tr' ++ tr).
  Proof.
    intros H H' H'' tr tr' steps_many steps_many'.
    induction steps_many'.
    simpl. auto.
    rewrite <- app_comm_cons. eauto.
  Qed.
End MemoryModelFacts.

Module ModelFacts (L : LANGUAGE_MODEL) (M : MEMORY_MODEL L)
  (Sys : SYSTEM_MODEL L M).
  Import L.
  Import M.
  Import Sys.

  Module LFacts := LanguageModelFacts L.
  Import LFacts.

  Hint Constructors heap_step_many program_step_many system_step
    system_step_many.

  Lemma system_step_many_alt : forall H P H' P' H'' P'' tr tr',
    system_step_many H P H' P' tr ->
    system_step_many H' P' H'' P'' tr' ->
    system_step_many H P H'' P'' (tr' ++ tr).
  Proof.
    intros H P H' P' H'' P'' tr tr' steps_many steps_many'.
    induction steps_many'.
    simpl. auto.
    rewrite <- app_comm_cons. eauto.
  Qed.

  Lemma system_step_many_one : forall H1 P1 H2 P2 e,
    system_step_many H1 P1 H2 P2 (e :: nil) ->
    system_step H1 P1 e H2 P2.
  Proof.
    intros H1 P1 H2 P2 e steps_many.
    inversion steps_many as [ | ? ? ? ? ? ? ? ? steps_many' ].
    inversion steps_many'.
    subst. auto.
  Qed.

  Lemma system_step_many_append : forall H1 P1 H2 P2 H3 P3 tr1 tr2,
    system_step_many H1 P1 H2 P2 tr1 ->
    system_step_many H2 P2 H3 P3 tr2 ->
    system_step_many H1 P1 H3 P3 (tr2 ++ tr1).
  Proof.
    intros H1 P1 H2 P2 H3 P3 tr1 tr2 steps_many1 steps_many2.
    induction steps_many2.
    simpl. auto.
    eauto.
    rewrite <- app_comm_cons. eauto.
  Qed.

  Lemma system_step_many_append_inversion : forall H1 P1 H2 P2 tr1 tr2,
    system_step_many H1 P1 H2 P2 (tr2 ++ tr1) ->
    exists H3, exists P3,
      system_step_many H1 P1 H3 P3 tr1 /\
      system_step_many H3 P3 H2 P2 tr2.
  Proof.
    intros H1 P1 H2 P2 tr1 tr2 steps_many.
    generalize dependent P2.
    generalize dependent H2.
    induction tr2 as [ | e tr2 ]; intros H2 P2 steps_many.
    simpl in steps_many. eauto.
    inversion steps_many as [ | ? ? ? ? ? ? ? ? steps_many' ]; subst.
    pose proof IHtr2 _ _ steps_many' as [ ? [ ? [ ? ? ] ] ].
    eauto.
  Qed.

  Lemma system_step_many_index_inversion :
    forall H P H' P' tr j e,
      system_step_many H P H' P' tr ->
      index tr j e ->
      exists H'', exists P'', exists H''', exists P''', exists tr', exists tr'',
        system_step_many H P H'' P'' tr'
        /\ system_step H'' P'' e H''' P'''
        /\ system_step_many H''' P''' H' P' tr''
        /\ tr = tr'' ++ (e :: nil) ++ tr'
        /\ length tr'' = j.
  Proof.
    intros H P H' P' tr j e steps_many idx.

    (* Proof by induction on the derivation of steps_many. *)
    generalize dependent j.
    induction steps_many as [ | H P H'' P'' H' P' tr e' steps_many ? steps ];
      intros j idx; subst.

    (* Zero steps total - vacuous. *)
    inversion idx.

    (* At least one step - was j the most recent? *)
    inversion idx as [ | ? j' ? ? idx' ]; subst.

    (* Yes - so tr' is nil and tr'' is tr. *)
    exists H''. exists P''. exists H'. exists P'. exists tr.
    exists (@nil (effect val)).
    simpl. auto.

    (* No - use the induction hypothesis. *)
    pose proof IHsteps_many _ idx'  as [ ? [ ? [ ? [ ? [ ? [ ? [
      steps_many' [ steps' [ steps_many'' [ ? ? ] ] ] ] ] ] ] ] ] ].
    subst. repeat esplit. apply steps_many'. eauto.
    eapply system_step_many_trans; eauto. eauto. eauto.
  Qed.

  Lemma system_step_many_index_inversion_2 :
    forall H1 P1 H2 P2 tr j1 e1 j2 e2,
      system_step_many H1 P1 H2 P2 tr ->
      index tr j1 e1 ->
      index tr j2 e2 ->
      j2 < j1 ->
      exists H3, exists P3, exists H4, exists P4, exists H5, exists P5,
        exists H6, exists P6, exists tr1, exists tr2, exists tr3,
          system_step_many H1 P1 H3 P3 tr1 /\
          system_step H3 P3 e1 H4 P4 /\
          system_step_many H4 P4 H5 P5 tr2 /\
          system_step H5 P5 e2 H6 P6 /\
          system_step_many H6 P6 H2 P2 tr3 /\
          tr = tr3 ++ (e2 :: nil) ++ tr2 ++ (e1 :: nil) ++ tr1 /\
          length tr3 = j2 /\
          length tr3 + length tr2 + 1 = j1.
  Proof.
    intros H1 P1 H2 P2 tr j1 e1 j2 e2 steps_many idx1 idx2 j2_lt.

    (* Proof by induction on system_step_many. *)
    generalize dependent j2.
    generalize dependent j1.
    induction steps_many as [ | H1 P1 H3 P3 H2 P2 tr e steps_many ? steps ];
      intros j1 idx1 j2 idx2 j2_lt.

    (* Zero steps - vacuous *)
    inversion idx1.

    (* Multiple steps - is j2 the first one? *)
    induction j2 as [ | j2 ].

    (* Yes - so we can infer the trace, noting that j1 must be at least 1. *)
    inversion idx2; subst.
    induction j1 as [ | j1 ].
    inversion j2_lt.
    inversion idx1 as [ | ? ? ? ? idx1' ]. subst.
    pose proof system_step_many_index_inversion steps_many idx1' as [ H4 [ P4
      [ H5 [ P5 [ tr1 [ tr2 [ ? [ ? [ ? [ ? ? ] ] ] ] ] ] ] ] ] ].
    exists H4. exists P4. exists H5. exists P5. exists H3. exists P3.
    exists H2. exists P2. exists tr1. exists tr2. exists (@nil (effect val)).
    subst tr j1. simpl. rewrite plus_1. auto 8.

    (* No - so apply the inductive hypothesis. *)
    induction j1 as [ | j1 ].
    inversion j2_lt.
    apply lt_S_n in j2_lt.
    inversion idx1 as [ | ? ? ? ? idx1' ]. subst.
    inversion idx2 as [ | ? ? ? ? idx2' ]. subst.
    clear IHj1 IHj2 idx1 idx2.
    pose proof IHsteps_many j1 idx1' j2 idx2' j2_lt as [ H4 [ P4 [ H5 [ P5
      [ H6 [ P6 [ H7 [ P7 [ tr1 [ tr2 [ tr3 [ ? [ ? [ ? [ ? [ ? [ ? [ ? ? ] ]
      ] ] ] ] ] ] ] ] ] ] ] ] ] ] ] ].
    exists H4. exists P4. exists H5. exists P5. exists H6. exists P6.
    exists H7. exists P7. exists tr1. exists tr2. exists (e :: tr3).
    subst tr j1. simpl. eauto 10.
  Qed.

  Lemma system_step_many_index_inversion_adjacent :
    forall H1 P1 H2 P2 tr j e1 e2,
      system_step_many H1 P1 H2 P2 tr ->
      index tr (j + 1) e1 ->
      index tr j e2 ->
      exists H3, exists P3, exists H4, exists P4, exists H5, exists P5,
        exists tr1, exists tr2,
          system_step_many H1 P1 H3 P3 tr1 /\
          system_step H3 P3 e1 H4 P4 /\
          system_step H4 P4 e2 H5 P5 /\
          system_step_many H5 P5 H2 P2 tr2 /\
          tr = tr2 ++ (e2 :: nil) ++ (e1 :: nil) ++ tr1 /\
          length tr2 = j.
  Proof.
    intros H1 P1 H2 P2 tr j e1 e2 steps_many idx1 idx2.
    assert (j < j + 1) as j_lt. omega.
    pose proof system_step_many_index_inversion_2 steps_many idx1 idx2 j_lt
      as [ ? [ ? [ ? [ ? [ ? [ ? [ ? [ ? [ tr1 [ tr2 [ tr3 [ ? [ ?
        [ steps_nil [ ? [ ? [ ? [ ? ? ] ] ] ] ] ] ] ] ] ] ] ] ] ] ] ] ] ].
    assert (tr2 = nil). rewrite (length_nil tr2). omega. subst tr2.
    inversion steps_nil; subst.
    repeat esplit; eauto.
  Qed.

  Lemma system_step_many_empty : forall H1 H2 tr P,
    heap_step_many H1 H2 tr ->
    erase_trace tr = nil ->
    system_step_many H1 P H2 P tr.
  Proof with auto.
    intros H1 H2 tr P StepMany Erase.
    induction StepMany...
    destruct e as [ [ ? ? ] | ].
    inversion Erase.
    simpl in Erase.
    eapply system_step_many_trans...
  Qed.

  Lemma system_step_many_mutual : forall H1 H2 P1 P2 tr t a,
    heap_step_many H1 H2 tr ->
    program_step P1 (Some (t, a)) P2 ->
    erase_trace tr = (t, a) :: nil ->
    system_step_many H1 P1 H2 P2 tr.
  Proof with eauto.
    intros H1 H2 P1 P2 tr t a StepMany Step Erase.
    induction StepMany.
    inversion Erase.
    destruct e as [ [ t' a' ] | ].
    inversion Erase. subst.
    eapply system_step_many_trans...
    apply system_step_many_empty...
    simpl in Erase...
  Qed.
End ModelFacts.
