Require Export List.
Require Export Metatheory.

Set Implicit Arguments.

Definition loc := var.
Definition tid := var.
Definition rgn := var.

Inductive prot_state :=
| pr   : prot_state
| unpr : prot_state
| ro   : prot_state
| tl   : tid -> prot_state.

Section Effects.
  Variable val : Set.

  Inductive action : Set :=
  | pure   : action
  | newRgn : rgn -> action
  | alloc  : loc -> rgn -> val -> action
  | read   : loc -> val -> action
  | write  : loc -> val -> action
  | spawn  : tid -> action
  | makeP  : rgn -> action
  | makeU  : rgn -> action
  | makeRO : rgn -> action
  | makeTL : rgn -> action
  | beginP : action
  | endP   : action.

  Definition effect := option (tid * action).

  Definition trace := list effect.

  Definition erased_trace := list (tid * action).

  Fixpoint erase_trace (tr : trace) : erased_trace :=
    match tr with
      | nil => nil
      | None :: tr' => erase_trace tr'
      | Some (t, a) :: tr' => (t, a) :: (erase_trace tr')
    end.

  Lemma app_erase : forall tr1 tr2,
    erase_trace (tr1 ++ tr2) = erase_trace tr1 ++ erase_trace tr2.
  Proof.
    intros tr1 tr2.
    induction tr1 as [ | [ [ ] | ] ]; simpl; f_equal; eauto.
  Qed.
End Effects.

Module Type LANGUAGE_MODEL.
  Parameter val : Set.
  Parameter program_state : Type.
  Parameter program_step : program_state -> effect val -> program_state -> Prop.

  Inductive program_step_many : program_state -> program_state -> trace val
    -> Prop :=
  | program_step_many_refl : forall P,
    program_step_many P P nil
  | program_step_many_trans : forall P1 P2 P3 tr e,
    program_step_many P1 P2 tr ->
    program_step P2 e P3 ->
    program_step_many P1 P3 (e :: tr).

  Axiom no_empty_steps : forall P P',
    ~ (program_step P None P').

  Parameter eq_program_state : program_state -> program_state -> Prop.
  Axiom eq_program_state_refl : forall P,
    eq_program_state P P.
  Axiom eq_program_state_sym : forall P1 P2,
    eq_program_state P1 P2 ->
    eq_program_state P2 P1.
  Axiom eq_program_state_trans : forall P1 P2 P3,
    eq_program_state P1 P2 ->
    eq_program_state P2 P3 ->
    eq_program_state P1 P3.
  Axiom eq_program_state_step : forall P1 e P2 P1',
    program_step P1 e P2 ->
    eq_program_state P1 P1' ->
    exists P2',
      eq_program_state P2 P2'
      /\ program_step P1' e P2'.
  Add Parametric Relation : program_state eq_program_state
    reflexivity proved by eq_program_state_refl
    symmetry proved by eq_program_state_sym
    transitivity proved by eq_program_state_trans
  as eq_program_state_rel.
End LANGUAGE_MODEL.

Module Type MEMORY_MODEL (L : LANGUAGE_MODEL).
  Parameter heap : Type.
  Parameter empty_heap : heap.
  Parameter heap_step : heap -> effect L.val -> heap -> Prop.

  Inductive heap_step_many : heap -> heap -> trace L.val -> Prop :=
  | heap_step_many_refl : forall H,
    heap_step_many H H nil
  | heap_step_many_trans : forall H1 H2 H3 tr e,
    heap_step_many H1 H2 tr ->
    heap_step H2 e H3 ->
    heap_step_many H1 H3 (e :: tr).
End MEMORY_MODEL.

Module Type SYSTEM_MODEL (L : LANGUAGE_MODEL) (M : MEMORY_MODEL L).
  Inductive system_step : M.heap -> L.program_state -> effect L.val
    -> M.heap -> L.program_state -> Prop :=
  | system_step_mutual : forall H1 H2 P1 P2 t a,
    L.program_step P1 (Some (t, a)) P2 ->
    M.heap_step H1 (Some (t, a)) H2 ->
    system_step H1 P1 (Some (t, a)) H2 P2
  | system_step_heap : forall H1 H2 P,
    M.heap_step H1 None H2 ->
    system_step H1 P None H2 P.

  Inductive system_step_many : M.heap -> L.program_state -> M.heap
    -> L.program_state -> trace L.val -> Prop :=
  | system_step_many_refl : forall H P,
    system_step_many H P H P nil
  | system_step_many_trans : forall H1 P1 H2 P2 H3 P3 tr e,
    system_step_many H1 P1 H2 P2 tr ->
    system_step H2 P2 e H3 P3 ->
    system_step_many H1 P1 H3 P3 (e :: tr).
End SYSTEM_MODEL.
