Require Import heaps.
Require Import maps.
Require Import models.

Set Implicit Arguments.

Module StrongLazy (L : LANGUAGE_MODEL) (Heap : HEAP L) <: MEMORY_MODEL L.
  Import Heap.

  Definition heap := weak_heap.
  Definition empty_heap := empty_weak_heap.

  Inductive step : heap -> effect L.val -> heap -> Prop :=
  | step_pure : forall ts S R TS t,
    strong_active ts t ->
    step (ts, S, R, TS) (Some (t, pure _)) (ts, S, R, TS)
  | step_newRgn : forall ts S R TS t r,
    strong_active ts t ->
    ~ VarMap.In r R ->
    step (ts, S, R, TS) (Some (t, newRgn _ r)) (ts, S, VarMap.add r pr R, TS)
  | step_alloc : forall ts S R TS t l r v,
    strong_active ts t ->
    ~ VarMap.In l S ->
    step (ts, S, R, TS) (Some (t, alloc l r v))
    (ts, VarMap.add l (r, v) S, R, TS)
  | step_tx_read1 : forall S R TS t l r v,
    VarMap.MapsTo l (r, v) S ->
    ~ VarMap.In l TS ->
    step (tx_running t, S, R, TS) (Some (t, read l v))
    (tx_running t, S, R, VarMap.add l (tx_ro, v) TS)
  | step_tx_read2 : forall S R TS t l s v,
    VarMap.MapsTo l (s, v) TS ->
    step (tx_running t, S, R, TS) (Some (t, read l v)) (tx_running t, S, R, TS)
  | step_read : forall S R TS t l r v,
    VarMap.MapsTo l (r, v) S ->
    step (tx_none, S, R, TS) (Some (t, read l v)) (tx_none, S, R, TS)
  | step_tx_write : forall S R TS t l v,
    step (tx_running t, S, R, TS) (Some (t, write l v))
    (tx_running t, S, R, VarMap.add l (tx_mod, v) TS)
  | step_write : forall S R TS t l r v v',
    VarMap.MapsTo l (r, v) S ->
    step (tx_none, S, R, TS) (Some (t, write l v'))
    (tx_none, VarMap.add l (r, v') S, R, TS)
  | step_spawn : forall S R TS t t',
    step (tx_none, S, R, TS) (Some (t, spawn _ t')) (tx_none, S, R, TS)
  | step_makeP : forall S R TS t r,
    step (tx_none, S, R, TS) (Some (t, makeP _ r))
    (tx_none, S, VarMap.add r pr R, TS)
  | step_makeU : forall S R TS t r,
    step (tx_none, S, R, TS) (Some (t, makeU _ r))
    (tx_none, S, VarMap.add r unpr R, TS)
  | step_makeRO : forall S R TS t r,
    step (tx_none, S, R, TS) (Some (t, makeRO _ r))
    (tx_none, S, VarMap.add r ro R, TS)
  | step_makeTL : forall S R TS t r,
    step (tx_none, S, R, TS) (Some (t, makeTL _ r))
    (tx_none, S, VarMap.add r (tl t) R, TS)
  | step_beginP : forall S R TS t,
    step (tx_none, S, R, TS) (Some (t, beginP _)) (tx_running t, S, R, TS)
  | step_endP : forall S R TS t,
    step (tx_running t, S, R, TS) (Some (t, endP _)) (tx_commit t, S, R, TS)
  | step_commit_ro : forall S R TS t l v,
    VarMap.MapsTo l (tx_ro, v) TS ->
    step (tx_commit t, S, R, TS) None (tx_commit t, S, R, VarMap.remove l TS)
  | step_commit_mod : forall S R TS t l r v v',
    VarMap.MapsTo l (r, v) S ->
    VarMap.MapsTo l (tx_mod, v') TS ->
    step (tx_commit t, S, R, TS) None
    (tx_commit t, VarMap.add l (r, v') S, R, VarMap.remove l TS)
  | step_end_commit : forall S R TS t,
    (forall l, ~ VarMap.In l TS) ->
    step (tx_commit t, S, R, TS) None (tx_none, S, R, TS).

  Definition heap_step := step.

  Inductive heap_step_many : heap -> heap -> trace L.val -> Prop :=
  | heap_step_many_refl : forall H,
    heap_step_many H H nil
  | heap_step_many_trans : forall H1 H2 H3 tr e,
    heap_step_many H1 H2 tr ->
    heap_step H2 e H3 ->
    heap_step_many H1 H3 (e :: tr).

  Lemma eq_weak_heap_step : forall H1 e H2 H1',
    heap_step H1 e H2 ->
    eq_weak_heap H1 H1' ->
    exists H2',
      eq_weak_heap H2 H2'
      /\ heap_step H1' e H2'.
  Proof with eauto.
    intros H1 e H2 H1' Step Eq.
    inversion Step; subst; inversion Eq; subst;
      esplit; rewrite and_comm; split.

    apply step_pure...
    auto.

    eapply step_newRgn...
    rewrite <- VarMapFacts.In_m...
    apply eq_weak_heap_def...
    apply VarMapFacts.add_m...

    apply step_alloc...
    rewrite <- VarMapFacts.In_m...
    apply eq_weak_heap_def...
    apply VarMapFacts.add_m...

    eapply step_tx_read1.
    rewrite <- VarMapFacts.MapsTo_m...
    rewrite <- VarMapFacts.In_m...
    apply eq_weak_heap_def...
    apply VarMapFacts.add_m...

    eapply step_tx_read2.
    rewrite <- VarMapFacts.MapsTo_m...
    auto.

    eapply step_read.
    rewrite <- VarMapFacts.MapsTo_m...
    auto.

    apply step_tx_write.
    apply eq_weak_heap_def...
    apply VarMapFacts.add_m...

    eapply step_write.
    rewrite <- VarMapFacts.MapsTo_m...
    apply eq_weak_heap_def...
    apply VarMapFacts.add_m...

    apply step_spawn.
    auto.

    apply step_makeP.
    apply eq_weak_heap_def...
    apply VarMapFacts.add_m...

    apply step_makeU.
    apply eq_weak_heap_def...
    apply VarMapFacts.add_m...

    apply step_makeRO.
    apply eq_weak_heap_def...
    apply VarMapFacts.add_m...

    apply step_makeTL.
    apply eq_weak_heap_def...
    apply VarMapFacts.add_m...

    apply step_beginP.
    apply eq_weak_heap_def...

    apply step_endP.
    apply eq_weak_heap_def...

    eapply step_commit_ro.
    rewrite <- VarMapFacts.MapsTo_m...
    apply eq_weak_heap_def...
    apply VarMapFacts.remove_m...

    eapply step_commit_mod.
    rewrite <- VarMapFacts.MapsTo_m...
    rewrite <- VarMapFacts.MapsTo_m...
    apply eq_weak_heap_def...
    apply VarMapFacts.add_m...
    apply VarMapFacts.remove_m...

    apply step_end_commit.
    intro. rewrite <- VarMapFacts.In_m...
    eapply eq_weak_heap_def...
  Qed.
End StrongLazy.

Ltac inversion_lazy_step Step t l r v v' Active RgnFresh Fresh Read TxFresh
  TxRead Empty :=
  inversion Step as [
    ? ? ? ? ? Active (* pure *)
    | ? ? ? ? ? r Active RgnFresh (* newRgn *)
    | ? ? ? ? ? l r v Active Fresh (* alloc *)
    | ? ? ? ? l r v Read TxFresh (* tx_read1 *)
    | ? ? ? ? l s v TxRead (* tx_read2 *)
    | ? ? ? ? l r v Read (* read *)
    | ? ? ? ? l v (* tx_write *)
    | ? ? ? ? l r v v' Read (* write *)
    | ? ? ? ? t (* spawn *)
    | ? ? ? ? r (* makeP *)
    | ? ? ? ? r (* makeU *)
    | ? ? ? ? r (* makeRO *)
    | ? ? ? ? r (* makeTL *)
    | (* beginP *)
    | (* endP *)
    | ? ? ? ? l v TxRead (* commit_ro *)
    | ? ? ? ? l r v v' Read TxRead (* commit_mod *)
    | ? ? ? ? Empty (* end_commit *)
  ]; subst.
