Require Import maps.
Require Import models.

Set Implicit Arguments.

Open Scope type_scope.

Module Type HEAP_TYPING.
  Parameter heap_typing : Set -> Type.
  Section Types.
    Variable typ : Set.
    Parameter empty_heap_typing : heap_typing typ.
    Parameter extends : heap_typing typ -> heap_typing typ -> Prop.
    Axiom extends_self : forall (HT : heap_typing typ),
      extends HT HT.
  End Types.
End HEAP_TYPING.

Module Type TYPE_MODEL.
  Declare Module L : LANGUAGE_MODEL.
  Declare Module HeapTyp : HEAP_TYPING.
  Parameter typ : Set.
  Parameter typing_val : HeapTyp.heap_typing typ -> L.val -> typ -> Prop.
  Parameter typing_prgm : HeapTyp.heap_typing typ -> L.program_state -> Prop.
  Axiom typing_stability : forall HT v T HT',
    typing_val HT v T ->
    HeapTyp.extends HT HT' ->
    typing_val HT' v T.
  Parameter well_typed_heap_action : HeapTyp.heap_typing typ -> action L.val
    -> Prop.
  Parameter well_typed_program_action : HeapTyp.heap_typing typ -> action L.val
    -> HeapTyp.heap_typing typ -> Prop.
  Axiom well_typed_extends :
    forall (HT : HeapTyp.heap_typing typ) a HT',
      well_typed_heap_action HT a ->
      well_typed_program_action HT a HT' ->
      HeapTyp.extends HT HT'.
End TYPE_MODEL.

Module Type TYPED_MEMORY_MODEL (Typ : TYPE_MODEL).
  Declare Module M : MEMORY_MODEL Typ.L.
  Parameter typing_heap : Typ.HeapTyp.heap_typing Typ.typ -> M.heap -> Prop.
  Axiom typing_empty_heap :
    typing_heap (Typ.HeapTyp.empty_heap_typing _) M.empty_heap.
End TYPED_MEMORY_MODEL.

Module Type PROGRAM_TYPE_PRESERVATION (Typ : TYPE_MODEL).
  Axiom type_preservation_program_step_mutual : forall HT P t a P',
    Typ.typing_prgm HT P ->
    Typ.L.program_step P (Some (t, a)) P' ->
    Typ.well_typed_heap_action HT a ->
    exists HT',
      Typ.well_typed_program_action HT a HT'
      /\ Typ.typing_prgm HT' P'.
End PROGRAM_TYPE_PRESERVATION.

Module Type HEAP_TYPE_PRESERVATION (Typ : TYPE_MODEL).
  Declare Module TypM : TYPED_MEMORY_MODEL Typ.

  Axiom well_typed_heap_step_mutual : forall HT H t a H',
    TypM.typing_heap HT H ->
    TypM.M.heap_step H (Some (t, a)) H' ->
    Typ.well_typed_heap_action HT a.

  Axiom type_preservation_heap_step_mutual : forall HT H t a H' HT',
    TypM.typing_heap HT H ->
    TypM.M.heap_step H (Some (t, a)) H' ->
    Typ.well_typed_program_action HT a HT' ->
    TypM.typing_heap HT' H'.

  Axiom type_preservation_heap_step_empty : forall HT H H',
    TypM.typing_heap HT H ->
    TypM.M.heap_step H None H' ->
    TypM.typing_heap HT H'.
End HEAP_TYPE_PRESERVATION.

Module Type_Preservation (HeapTyp : HEAP_TYPING)
  (Typ : TYPE_MODEL with Module HeapTyp := HeapTyp)
  (PrgmPres : PROGRAM_TYPE_PRESERVATION Typ)
  (HeapPres : HEAP_TYPE_PRESERVATION Typ)
  (Sys : SYSTEM_MODEL Typ.L HeapPres.TypM.M).

  Theorem type_preservation : forall HT H P e H' P',
    HeapPres.TypM.typing_heap HT H ->
    Typ.typing_prgm HT P ->
    Sys.system_step H P e H' P' ->
    exists HT',
      HeapTyp.extends HT HT'
      /\ HeapPres.TypM.typing_heap HT' H'
      /\ Typ.typing_prgm HT' P'.
  Proof with eauto.
    intros ? ? ? ? ? ? ? ? Step.
    inversion Step as [ ? ? ? ? t a PrgmStep HeapStep
      | ? ? ? HeapStep ]; subst.
    forwards : HeapPres.well_typed_heap_step_mutual...
    forwards [ ? [ ? ? ] ] : PrgmPres.type_preservation_program_step_mutual...
    forwards : HeapPres.type_preservation_heap_step_mutual...
    forwards : Typ.well_typed_extends...
    forwards : HeapPres.type_preservation_heap_step_empty...
    forwards : HeapTyp.extends_self...
  Qed.

  Corollary type_preservation_many : forall P H' P' tr,
    Typ.typing_prgm (HeapTyp.empty_heap_typing _) P ->
    Sys.system_step_many HeapPres.TypM.M.empty_heap P H' P' tr ->
    exists HT,
      HeapPres.TypM.typing_heap HT H'
      /\ Typ.typing_prgm HT P'.
  Proof with eauto.
    intros ? ? ? ? TypingPrgm StepMany.
    remember (HeapTyp.empty_heap_typing Typ.typ) as HT.
    remember HeapPres.TypM.M.empty_heap as H.
    induction StepMany; subst.
    forwards : HeapPres.TypM.typing_empty_heap...
    forwards [ ? [ ? ? ] ] : IHStepMany (>>> __ __)...
    forwards [ ? [ ? [ ? ? ] ] ] : type_preservation...
  Qed.
End Type_Preservation.

Module HeapTyping <: HEAP_TYPING.
  Section Types.
    Variable typ : Set.

    Definition prot_typing := VarMap.t unit.
    Definition store_typing := VarMap.t typ.
    Definition heap_typing := prot_typing * store_typing.

    Definition empty_heap_typing : heap_typing :=
      (VarMap.empty _, VarMap.empty _).

    Inductive extends_ind : heap_typing -> heap_typing -> Prop :=
    | extends_def : forall PT ST PT' ST',
      (forall r, VarMap.In r PT -> VarMap.In r PT') ->
      (forall l T, VarMap.MapsTo l T ST -> VarMap.MapsTo l T ST') ->
      extends_ind (PT, ST) (PT', ST').

    Definition extends := extends_ind.

    Lemma extends_self : forall HT,
      extends HT HT.
    Proof with auto.
      intros [ ? ? ].
      apply extends_def...
    Qed.
  End Types.
End HeapTyping.
